sections = [
    "title",
    "options",
    "filenames",
    "weights_coefficients",
    "convergence",
    "output",
    "elements",
    "components",
    "minerals",
    "aqueous_species",
    "surface_complexes",
    "exchange_species",
    "default",
    "zones",
    "convergence_bounds",
    "end",
    "end_comments",
]
